
# Modul: edit.tcl
#
# Funktionen zum ndern der abstrakten Struktur.
#
# Das Modul bernimmt folgende Aufgaben:
#
# * Die nderungsfunktionen ndern nicht nur die
#   Sicht-Struktur, sondern ggf. auch die
#   Semantische Struktur.
#
# * Falls fr bestimmte Klassen Editierhooks definiert
#   sind, werden diese aufgerufen.
#
#
#
# Kette der nderungsfunktionen:
#
# edit -> change -> d -> c
#
# - Alle nerdungsoperationen, die nicht explizit die
#   Trennung zwischen semantischem Modell und Programmmodell
#   beachten, sollten nur "edit" verwenden.
#
# - nderungsoperationen, die explizit die semantische
#   oder die Programmstruktur ndern sollten "change"
#   benutzen.
#

namespace eval ::edit {}
namespace eval ::edithook {}

# Hooks:
#
# * create_CLASS
# * setAttr_CLASS
#
# * initValue



# --- internal functions --------------------------------------------------------------


proc edit::ask {question} {
    return [CDlg::Question $question]
}


set edit::newObjects ""


proc edit::doContextSensitiveInit {} {

    set todo $::edit::newObjects
    set ::edit::newObjects ""
    # exernal hooks could create other new objects!

    foreach obj $todo {
        assignInitialName $obj
        executeCreateHooks $obj
    }

}


proc edit::assignInitialName {obj} {
    if [c::getSubElement $obj name -var nameAttr] {
        if [empty [c::getValue $nameAttr]] {
            change::setValue $nameAttr [basiccomp::computeInitialName $obj]
        }
    }
}


proc edit::executeCreateHooks {obj} {
    set class [c::getClassName $obj]
    if [procExists "edithook::create_$class"] {
        edithook::create_$class $obj
    }    
}


proc edit::tryInitExclamationmarkAttr {attr} {
    if { ![c::isAggregate $attr] || ![c::minMultiplicityIsOne $attr] } {
        return
    }
    set class [c::getNameOfType [c::getAcceptedType $attr]]
    set obj [createObject $class [getFile $attr]]
    edit::insert [c::icBelow $attr] $obj
}


proc edit::tryInitValAttr {attr} {
    foreach context [c::getContexts $attr] {
        if [info exists edithook::initValue($context)] {
            change::setValue $attr $edithook::initValue($context)
            return
        }
    }
}

proc edit::doNonContextSensitiveInit {obj} {
    foreach attr [c::getSubElements $obj] {
        tryInitValAttr $attr
        tryInitExclamationmarkAttr $attr
    }
}

# --- clipboard -----------------------------------------------------------------------


proc edit::cutIntoClipboard {objNode} {
    clipboard::cut $objNode
}

proc edit::mark {objNode} {
    clipboard::mark $objNode
}

proc edit::copyIntoClipboard {objNode} {
    clipboard::copy $objNode
}


# --- edit functions -------------------------------------------------------------------



proc edit::createRoot {} {
    set obj [change::createRoot]
    doNonContextSensitiveInit $obj
    lappend ::edit::newObjects $obj
    return $obj
}


proc edit::createObject {class file} {
    set obj [change::createObject $class $file]
    doNonContextSensitiveInit $obj
    lappend ::edit::newObjects $obj
    return $obj
}


proc edit::createModelElement {rep modelElementVar} {
    # erzeugt ein Modellelement fuer Repraensentant $rep.
    # Liefert true, wenn das mglich ist.
    upvar $modelElementVar modelElement

    set modelClass [c::getModelClass [c::getClassName $rep]]
    if [procExists c::${modelClass}_new] {
        set modelElement [createObject $modelClass $file]
        setModel $rep $modelElement
        return 1
    } else {
        return 0
    }
}


proc edit::insert {refObj newElem} {

    if [equal $refObj $newElem] {
        # just ignore this case.
        return
    }

    change::insert $refObj $newElem

    if [c::getModel [c::getParent $refObj] aggModel] {

        # -> Repraesentanten-Liste
        # Auch das Modell muss noch geaendert werden!
        # Defaultverhalten: Aendere auch das Modell

        # finde entsprechendes Modellelement oder erzeuge ein neues
        if [c::getModel $newElem modelElem] {
            # es existiert schon ein Modellelement;
        } else {
            if ![createModelElement $newElem modelElem] {
                # dies ist ein nur-sicht element.
                # nichts tun!
                return
            }
        }

        # suche Stelle, Einfuegestelle im Modell -> refObj
        while ![equal [c::get $refObj.SEMANTIC.PARENT] $aggModel] {
            set refObj [c::getSuccessor $refObj]
        }

        # einfuegen
        change::insert [c::get $refObj.SEMANTIC] $modelElem]
    }

}



proc edit::setValue {obj value} {
    foreach context [c::getContexts $obj] {
        if [procExists "edithook::setAttr_$context]"] {
            edithook::setAttr_$context $obj
        }
    }
    change::setValue $obj $value
    if [equal [c::getRole $obj] modelRef] {
        return
    }
    if [c::getModel $obj model] {
        change::setValue $model $value
    }
}


proc edit::delete {obj} {
    if ![canExtract $obj] {
        CDlg::Error [i18n::text "Can't delete this object!"] "Delete"
        return
    }
    if [c::getModel $obj model] {
        switch [ask [i18n::text "Delete model element as well?"]] {
            yes {                
                change::delete $model
                change::delete $obj
            }
            no {
                change::delete $obj
            }
            cancel {
                return
            }
        }
    } else {
        change::delete $obj
    }
}


# --- convenient edit functions ---------------------------------------------------------------


proc edit::setAttr {obj attr value} {
    set sub [c::getSubElement $obj $attr]
    setValue $sub $value
}


proc edit::insertConv {obj attr elem} {
    change::insert [c::getSubElement $obj $attr] $elem
    return $elem
}


proc edit::setModel {objNode value} {
    change::setValue [getSubElement $objNode modelRef] $value
}


proc edit::insertNew {aggNode class} {
    insert $aggNode [createObject $class [getFile $aggNode]]
}


proc edit::setClassOfChild {aggNode newClass} {
    if [equal [c::getClassOfChild $aggNode] $newClass] {
        return
    }
    set children [c::getList $aggNode.CHILDREN]
    if ![empty $children] {
        delete [lindex $children 0]
    }
    if ![empty $newClass] {
        insertNew [c::icBelow $aggNode] $newClass
    }
}


# --- check functions ------------------------------------------------------------------------


proc edit::getAcceptedType {insContext} {
    return [c::getAcceptedType [c::getParent $insContext]]
}


proc edit::canInsert {insertionContext newElem} {
    return [c::canInsert $insertionContext $newElem]
}


proc edit::canExtract {obj} {
    assert { [c::isLanguageObject $obj] }
    return [expr { [c::canExtract $obj] && ![c::minMultiplicityIsOne [c::getParent $obj]] }]
}


# --- other functions ------------------------------------------------------------------------


proc edit::forceUpdate {obj} {
    file::forceUpdate [file::getFileOfNode $obj]
}


proc edit::getFile {node} {
    return [change::getFile $node]
}

